#ifndef NPSTAT_LORPE1D_HH_
#define NPSTAT_LORPE1D_HH_

/*!
// \file LOrPE1D.hh
//
// \brief Convenience class which aggregates the kernel and the data
//        for unbinned density estimation by LOrPE
//
// Author: I. Volobouev
//
// May 2022
*/

#include <vector>
#include <cassert>
#include <stdexcept>
#include <algorithm>

#include "npstat/stat/LOrPE1DSymbetaKernel.hh"
#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    template <class Numeric>
    class LOrPE1D
    {
    public:
        inline LOrPE1D(const LOrPE1DSymbetaKernel& i_kernel,
                       const Numeric* i_coords, const unsigned long i_nCoords)
            : kernel_(i_kernel), coords_(i_coords, i_coords + i_nCoords)
        {
            if (coords_.empty()) throw std::invalid_argument(
                "In npstat::LOrPE1D constructor: empty point sample");
            assert(i_coords);
            sortAndNormalize();
        }

        inline LOrPE1D(const LOrPE1DSymbetaKernel& i_kernel,
                       const std::vector<Numeric>& i_coords)
            : kernel_(i_kernel), coords_(i_coords)
        {
            if (coords_.empty()) throw std::invalid_argument(
                "In npstat::LOrPE1D constructor: empty point sample");
            sortAndNormalize();
        }

        inline ~LOrPE1D() {}

        inline void setNormFactor(const double d) {kernel_.setNormFactor(d);}

        inline void setSample(const Numeric* i_coords,
                              const unsigned long i_nCoords)
        {
            if (!i_nCoords) throw std::invalid_argument(
                "In npstat::LOrPE1D::setSample: empty point sample");
            assert(i_coords);
            coords_.clear();
            coords_.reserve(i_nCoords);
            std::copy(i_coords, i_coords+i_nCoords, std::back_inserter(coords_));
            sortAndNormalize();
        }

        inline void setSample(const std::vector<Numeric>& i_coords)
        {
            if (i_coords.empty()) throw std::invalid_argument(
                "In npstat::LOrPE1D::setSample: empty point sample");
            coords_ = i_coords;
            sortAndNormalize();
        }

        inline const LOrPE1DSymbetaKernel& kernel() const {return kernel_;}
        inline double normFactor() const {return kernel_.normFactor();}
        inline const std::vector<Numeric>& coords() const {return coords_;}
        inline unsigned long nCoords() const {return coords_.size();}
        inline Numeric minCoordinate() const {return coords_[0];}
        inline Numeric maxCoordinate() const {return coords_.back();}

        inline double density(const double x, const double bw) const
            {return kernel_.lorpe(x, bw, &coords_[0], coords_.size(), true);}

        /** 
        // The lifetime of the lightweight functor returned by this
        // method must not exceed the lifetime of this object.
        //
        // Fcn can either be a functor which calculates the bandwidth
        // as a function of x or just a double (for the constant
        // bandwidth).
        */
        template <typename Fcn>
        inline LOrPE1DFunctorHelper<Numeric,Fcn> densityFunctor(
            Fcn bandwidthFunctor) const
        {
            return LOrPE1DFunctorHelper<Numeric,Fcn>(
                kernel_, bandwidthFunctor, &coords_[0], coords_.size(), true);
        }

        template <class Fcn>
        double densityIntegral(
            unsigned nIntegIntervals, unsigned nIntegPoints,
            Fcn bandwidthFunctor) const;

        template <class Fcn>
        double integratedSquaredError(
            const AbsDistribution1D& distro,
            const unsigned nIntegIntervals, const unsigned nIntegPoints,
            Fcn bandwidthFunctor) const;

    private:
        inline void sortAndNormalize()
        {
            std::sort(coords_.begin(), coords_.end());
            const double w = Private::SampleWeight<Numeric>(coords_).value;
            if (w <= 0.0) throw std::invalid_argument(
                "In npstat::LOrPE1D::sortAndNormalize: "
                "total sample weight must be positive");
            kernel_.setNormFactor(1.0/w);
        }

        LOrPE1DSymbetaKernel kernel_;
        std::vector<Numeric> coords_;
    };
}

#include "npstat/stat/LOrPE1D.icc"

#endif // NPSTAT_LORPE1D_HH_
